#include-once

Global Const $MAX_COLUMNS = 256
Global Const $MAX_BUFFER_SIZE = 65534
Global Enum $PLUGIN_HANDLE, $PLUGIN_COLUMNS, $WDX_PLUGIN_NAME
Global Enum $COLUMN_TYPE, $COLUMN_NAME, $COLUMN_UNIT, $COLUMN_MAXINDEX
Global Enum $ERROR_ERROPEN = 1, $ERROR_ERRHANDLE, $ERROR_ERRZEROCOLUMNS, $ERROR_ERRCOLUMN, $ERROR_ERRUNIT, $ERROR_ERRCALL, $ERROR_ERRVALUE

; $ERROR_ERROPEN - error openinh of DLL
; $ERROR_ERRHANDLE - errorous "handle"
; $ERROR_ERRZEROCOLUMNS - no columns was retrieved
; $ERROR_ERRCOLUMN - error in column name/index
; $ERROR_ERRUNIT - error in unit name/index
; $ERROR_ERRCALL - error calling function (usually absent function)
; $ERROR_ERRVALUE - error retrieving value (wrong or unsupported type)

Global Const $ft_nomorefields = 0 
Global Const $ft_numeric_32 = 1 
Global Const $ft_numeric_64 = 2 
Global Const $ft_numeric_floating = 3 
Global Const $ft_date = 4 
Global Const $ft_time = 5 
Global Const $ft_boolean = 6 
Global Const $ft_multiplechoice = 7 
; $ft_string conflicts with WinAPI constant
Global Const $ft_stringA = 8
Global Const $ft_fulltext = 9 
Global Const $ft_datetime = 10 
Global Const $ft_stringW = 11 

Global Const $ft_nosuchfield = -1
Global Const $ft_fileerror = -2
Global Const $ft_fieldempty = -3
Global Const $ft_ondemand = -4

Global Const $tagContentDefaultParamStruct = "int;dword;dword;char[256]"

; Result: special "handle" (acually array: [plugin handle, fields array])
; Error: 0 and sets @error
Func _WDXPluginOpen($sPlugin)
    Local $hPlugin = DllOpen($sPlugin)
    If @Error Then Return SetError($ERROR_ERROPEN, 0, 0)

    Local $pDPS = DllStructCreate($tagContentDefaultParamStruct)
    ; Struct size
    DllStructSetData($pDPS, 1, DllStructGetSize($pDPS))
    ; Plugin version 2.11
    DllStructSetData($pDPS, 2, 11)
    DllStructSetData($pDPS, 3, 2)
    ; "Recommended" settings file path
    DllStructSetData($pDPS, 4, EnvGet("COMMANDER_PATH") & "\contplug.ini")
    DllCall($hPlugin, "none", "ContentSetDefaultParams", "ptr", DllStructGetPtr($pDPS))

    ; Get columns
    Local $aColumns = __WDXGetColumnsInternal($hPlugin)
    If @error Then Return SetError(@error, 0, 0)

    Local $sPlugName = StringRegExpReplace($sPlugin, "(?m)^.+?\\([^\\]+)\.[^\.]+$", "\1")
    Local $aPlugin[3] = [$hPlugin, $aColumns, $sPlugName]
    Return $aPlugin
EndFunc

; Result: none
Func _WDXPluginClose(ByRef $aPlugin)
    If Not _WDXIsValidHandle($aPlugin) Then Return

    DllCall($aPlugin[$PLUGIN_HANDLE], "none", "ContentPluginUnloading")
    DllClose($aPlugin[$PLUGIN_HANDLE])
    $aPlugin = 0
EndFunc

; Result: 2D array of all columns: [type, name, units]
; Error: empty string and sets @error
Func _WDXGetColumns(ByRef $aPlugin)
    If Not _WDXIsValidHandle($aPlugin) Then Return SetError($ERROR_ERRHANDLE, 0, "")

    Return $aPlugin[$PLUGIN_COLUMNS]
EndFunc

; Return: zero-based column or unit index, 
; Error: -1 and sets @error
Func _WDXGetIndexByName(ByRef $aPlugin, $vColumn, $vUnit = Default)
    If Not _WDXIsValidHandle($aPlugin) Then Return SetError($ERROR_ERRHANDLE, 0, "")
    
    Local $aColumns = $aPlugin[$PLUGIN_COLUMNS]

    ; if field is given by name
    If Not StringIsDigit($vColumn) Then 
        $vColumn = __WDXGetColumnIndexByNameInternal($aColumns, $vColumn)
        If @error Then Return SetError(@error, 0, "")
    Else
        If Not (($vColumn >= 0) And $vColumn < UBound($aColumns)) Then Return SetError($ERROR_ERRCOLUMN, 0, -1)
    EndIf

    If $vUnit = Default Then Return $vColumn

    Local $aUnits = $aColumns[$vColumn][$COLUMN_UNIT]

    ; if unit is given by name
    If Not StringIsDigit($vUnit) Then 
        $vUnit = __WDXGetUnitIndexByNameInternal($aUnits, $vUnit)
        Return SetError(@error, 0, $vUnit)
    Else
        If Not (($vUnit >= 0) And $vUnit < UBound($aUnits)) Then Return SetError($ERROR_ERRUNIT, 0, -1)
    EndIf

    Return $vUnit
EndFunc

; Return: column value
; Error: empty string and sets @error
Func _WDXGetValue(ByRef $aPlugin, $sFileName, $vColumn, $vUnit = 0)
    If Not _WDXIsValidHandle($aPlugin) Then Return SetError($ERROR_ERRHANDLE, 0, "")

    Local $aColumns = $aPlugin[$PLUGIN_COLUMNS]

    ; if field is given by name
    If Not StringIsDigit($vColumn) Then 
        $vColumn = __WDXGetColumnIndexByNameInternal($aColumns, $vColumn)
        If @error Then Return SetError(@error, 0, "")
    Else
        If Not (($vColumn >= 0) And $vColumn < UBound($aColumns)) Then Return SetError($ERROR_ERRCOLUMN, 0, "")
    EndIf

    Local $aUnits = $aColumns[$vColumn][$COLUMN_UNIT]

    ; if unit is given by name
    If Not StringIsDigit($vUnit) Then 
        $vUnit = __WDXGetUnitIndexByNameInternal($aUnits, $vUnit)
        If @error Then Return SetError(@error, 0, "")
    Else
        If $vUnit = 0 Then
            ; allowed, do nothing
        ElseIf Not (($vUnit > 0) And $vUnit < UBound($aUnits)) Then 
            Return SetError($ERROR_ERRUNIT, 0, "")
        Endif
    EndIf

    Local $vFieldValue = __WDXGetFieldValue($aPlugin[$PLUGIN_HANDLE], $sFileName, $vColumn, $vUnit)
    Return SetError(@error, 0, $vFieldValue)
EndFunc

; Return: array of all values
; Error: empty string and sets @error
Func _WDXGetValues(ByRef $aPlugin, $sFileName)
    If Not _WDXIsValidHandle($aPlugin) Then Return SetError($ERROR_ERRHANDLE, 0, "")

    Local $aColumns = $aPlugin[$PLUGIN_COLUMNS]

    Local $aFieldValues[UBound($aColumns)]
    For $i = 0 To UBound($aColumns) - 1
        $vFieldValue = __WDXGetFieldValue($aPlugin[$PLUGIN_HANDLE], $sFileName, $i, 0)
        If @error Then ExitLoop
        $aFieldValues[$i] = $vFieldValue
    Next

    Return $aFieldValues
EndFunc

Func _WDXIsValidHandle(ByRef $aPlugin)
    Return IsArray($aPlugin) And ($aPlugin[$PLUGIN_HANDLE] <> 0) And IsArray($aPlugin[$PLUGIN_COLUMNS])
EndFunc

#cs --------------------------------------------------------------------------------------------------------------------------
Internal usage only
#ce --------------------------------------------------------------------------------------------------------------------------

Func __WDXGetColumnIndexByNameInternal(ByRef $aColumns, $sColumnName)
    For $i = 0 To UBound($aColumns) - 1 
        If $sColumnName = $aColumns[$i][$COLUMN_NAME] Then Return $i
    Next
    Return SetError($ERROR_ERRCOLUMN, 0, -1)
EndFunc

Func __WDXGetUnitIndexByNameInternal(ByRef $aUnits, $sUnitName)
    For $j = 0 To UBound($aUnits) - 1
        If $aUnits[$j] = $sUnitName Then Return $j
    Next
    Return SetError($ERROR_ERRUNIT, 0, -1)
EndFunc

Func __WDXGetColumnsInternal($hPlugin)
    Local $aRet
    Local $aColumns[$MAX_COLUMNS][$COLUMN_MAXINDEX], $nColumnCount = 0
    For $i = 0 To $MAX_COLUMNS - 1 
        $aRet = DllCall($hPlugin, "int", "ContentGetSupportedField", "int", $i, "str", "", "str", "", "int", $MAX_BUFFER_SIZE) 
        If @error Then Return SetError($ERROR_ERRCALL, 0, 0)
        ; checking also for negative values due some wrong plugins
        If $aRet[0] = $ft_nomorefields Or $aRet[0] < 0 Then ExitLoop
        $aColumns[$i][$COLUMN_TYPE] = $aRet[0]
        $aColumns[$i][$COLUMN_NAME] = $aRet[2]
        If $aRet[3] = "" Then
            $aColumns[$i][$COLUMN_UNIT] = $aRet[3]
        Else
            $aColumns[$i][$COLUMN_UNIT] = StringSplit($aRet[3], "|", 2)
        EndIf
        $nColumnCount += 1
    Next
    If $nColumnCount = 0 Then Return SetError($ERROR_ERRZEROCOLUMNS, 0, 0)
    ReDim $aColumns[$nColumnCount][$COLUMN_MAXINDEX]
    Return $aColumns
EndFunc

Func __WDXGetFieldValue($hPlugin, $sFileName, $nFieldIdx, $nUnitIdx)
    Local $pBuffer = DllStructCreate("byte[" & $MAX_BUFFER_SIZE & "]")
    $aRet = DllCall($hPlugin, "int", "ContentGetValueW", "wstr", $sFileName, "int", $nFieldIdx, "int", $nUnitIdx, "ptr", DllStructGetPtr($pBuffer), "int", $MAX_BUFFER_SIZE, "int", 0)
    ; in case of error "function not found", try to load ANSI version
    If @error = 3 Then $aRet = DllCall($hPlugin, "int", "ContentGetValue", "str", $sFileName, "int", $nFieldIdx, "int", $nUnitIdx, "ptr", DllStructGetPtr($pBuffer), "int", $MAX_BUFFER_SIZE, "int", 0) 
    If @error Then Return SetError($ERROR_ERRCALL, 0, "")
    Switch $aRet[0]
        Case $ft_numeric_32 
            $pValue = DllStructCreate("int", DllStructGetPtr($pBuffer)) 
        Case $ft_numeric_64 
            $pValue = DllStructCreate("int64", DllStructGetPtr($pBuffer)) 
        Case $ft_numeric_floating 
            $pValue = DllStructCreate("double", DllStructGetPtr($pBuffer)) 
        Case $ft_stringa, $ft_fulltext 
            $pValue = DllStructCreate("char[" & $MAX_BUFFER_SIZE & "]", DllStructGetPtr($pBuffer)) 
        Case $ft_stringw 
            $pValue = DllStructCreate("wchar[" & Int($MAX_BUFFER_SIZE / 2) & "]", DllStructGetPtr($pBuffer))
        Case $ft_date
            $pDate = DllStructCreate("word;word;word", DllStructGetPtr($pBuffer))
            Return StringFormat("%04d-%02d-%02d", _ 
                                DllStructGetData($pDate, 1), _ 
                                DllStructGetData($pDate, 2), _ 
                                DllStructGetData($pDate, 3))
        Case $ft_time
            $pTime = DllStructCreate("word;word;word", DllStructGetPtr($pBuffer))
            Return StringFormat("%02d:%02d:%02d", _ 
                                DllStructGetData($pTime, 1), _ 
                                DllStructGetData($pTime, 2), _ 
                                DllStructGetData($pTime, 3))
        Case $ft_datetime
            $pFileTime = DllStructCreate("dword;dword", DllStructGetPtr($pBuffer)) 
            $pLocalTime = DllStructCreate("dword;dword") 
            DllCall("Kernel32.dll", "int", "FileTimeToLocalFileTime", "ptr", DllStructGetPtr($pFileTime), "ptr", DllStructGetPtr($pLocalTime)) 
            $pValue = DllStructCreate("short;short;short;short;short;short;short;short") 
            DllCall("Kernel32.dll", "int", "FileTimeToSystemTime", "ptr", DllStructGetPtr($pLocalTime), "ptr", DllStructGetPtr($pValue)) 
            Return StringFormat("%04d-%02d-%02d %02d:%02d:%02d", _ 
                                DllStructGetData($pValue, 1), _ 
                                DllStructGetData($pValue, 2), _ 
                                DllStructGetData($pValue, 4), _ 
                                DllStructGetData($pValue, 5), _ 
                                DllStructGetData($pValue, 6), _ 
                                DllStructGetData($pValue, 7))
        Case $ft_boolean
            $pValue = DllStructCreate("boolean", DllStructGetPtr($pBuffer))
        Case $ft_nosuchfield, $ft_fileerror, $ft_fieldempty
            Return SetError($ERROR_ERRVALUE, 0, "")
        Case Else
            Return SetError($ERROR_ERRVALUE, 0, "")
    EndSwitch 
    Return DllStructGetData($pValue, 1)
EndFunc

