#include <TCWDXPlugin.au3>

; Need to be compiled with "Console" option enabled (not available in latest Autoit versions)
; Compiled with Autoit 3.3.8.1

; 2022-09-29 - Now always full path to testfile is passed to plugins
; 2021-09-21 - Fixed extra zero at the end of output, updated WDX library
; 2021-07-21 - Fixed writing to StdOut, added switch for setting output codepage
; 2020-12-19 - First release was not actually Unicode, interface improvements
; 2016-06-16 - Initial release

#NoTrayIcon

If $CmdLine[0] = 0 Then
    ConsoleWrite("No any parameters passed." & @CRLF & @CRLF & "Usage: " & @ScriptName & " [/C:codepage] <plugin.wdx> [testfile [column [unit]]]" & @CRLF)
    Exit
EndIf

; Handle switches
Global $iCodepage = 65001
Global $aParams[1] = [0]
For $i = 1 To $CmdLine[0]
    If StringLeft($CmdLine[$i], 3) = "/C:" Then
        $sCP = StringTrimLeft($CmdLine[$i], 3)
        If StringIsDigit($sCP) Then
            $iCodepage = Number($sCP)
            ContinueLoop
        Else
            ConsoleWrite("Wrong codepage option.")
            Exit
        EndIf
    EndIf
    $aParams[0] += 1
    ReDim $aParams[$aParams[0] + 1]
    $aParams[$aParams[0]] = $CmdLine[$i]
Next

If $aParams[0] = 0 Then
    ConsoleWrite("No required parameters passed.")
    Exit
EndIf

Global $aTypes[12] = ["Unknown", "Int32", "Int64", "Float", "Date", "Time", "Boolean", "MultipleChoice", "String", "FullText", "DateTime", "UnicodeString"]
Global $sPluginName = $aParams[1]
Global $sFileName = ""
Global $nColumnIdx = 0
Global $nUnitIdx = 0
Global $sFormatStr

If Not FileExists($sPluginName) Then
    ConsoleWrite("Plugin file does not exists.")
    Exit
EndIf

Global $aPlugin = _WDXPluginOpen($sPluginName)
If @error Then
    ConsoleWrite("Error opening plugin file, probably not WDX plugin.")
    Exit
EndIf

Opt("OnExitFunc", "ExitFunc")
Global $g_hKernel32 = DllOpen("kernel32.dll")
Global $g_hOutHandle = _GetStdHandle(-11)

DllCall("kernel32.dll", "bool", "SetConsoleOutputCP", "uint", $iCodepage)
;DllCall("kernel32.dll", "uint", "GetConsoleOutputCP")

If $aParams[0] = 1 Then
    $aColumns = _WDXGetColumns($aPlugin)
    ConsoleWrite("All columns: " & @CRLF & @CRLF)
    $nColPadSize = StringLen(UBound($aColumns))
    For $i = 0 To UBound($aColumns) - 1
        $sFormatStr = StringFormat("[%" & $nColPadSize & "s] -> %s (%s)\r\n", $i, $aColumns[$i][$COLUMN_NAME], _GetTypeString($aColumns[$i][$COLUMN_TYPE]))
        ConsoleWrite($sFormatStr)
        If IsArray($aColumns[$i][$COLUMN_UNIT]) Then
            $aUnits = $aColumns[$i][$COLUMN_UNIT]
            $nUnitPadSize = StringLen(UBound($aUnits))
            For $j = 0 To UBound($aUnits) - 1
                $sFormatStr = StringFormat("%" & $nColPadSize + 2 & "s [%" & $nUnitPadSize & "s] -> %s\r\n", "", $j, $aUnits[$j])
                ConsoleWrite($sFormatStr)
            Next
        EndIf
    Next
    Exit
EndIf

If $aParams[0] >= 2 Then
    ; some plugins fail if no full path to file passed (TC plugin API always passing full path)
    $sFileName = _GetFullPath($aParams[2])
    If Not FileExists($sFileName) Then
        ConsoleWrite("Test file does not exists.")
        Exit
    EndIf
EndIf

If $aParams[0] = 2 Then
    $aColumns = _WDXGetColumns($aPlugin)
    ConsoleWrite("All column values: " & @CRLF & @CRLF)
    For $i = 0 To UBound($aColumns) - 1
        $aUnits = $aColumns[$i][$COLUMN_UNIT]
        If IsArray($aUnits) Then
            ConsoleWrite($aColumns[$i][$COLUMN_NAME] & @CRLF)
            For $j = 0 To UBound($aUnits) - 1
                ConsoleWrite("    ")
                $sValue = _WDXGetValue($aPlugin, $sFileName, $i, $j)
                ConsoleWrite(_TextAlign($aUnits[$j], 28) & ": ")
                _WriteOutput($g_hOutHandle, $sValue, $iCodepage)
                ConsoleWrite(@CRLF)
            Next
        Else
            $sValue = _WDXGetValue($aPlugin, $sFileName, $i)
            ConsoleWrite(_TextAlign($aColumns[$i][$COLUMN_NAME], 32) & ": ")
            _WriteOutput($g_hOutHandle, $sValue, $iCodepage)
            ConsoleWrite(@CRLF)
        EndIf
    Next
    Exit
EndIf

If $aParams[0] >= 3 Then $nColumnIdx = $aParams[3]
If $aParams[0] = 4 Then $nUnitIdx = $aParams[4]

$sValue = _WDXGetValue($aPlugin, $sFileName, $nColumnIdx, $nUnitIdx)
If @error Then
    Switch @error
        Case $ERROR_ERRCOLUMN
            ConsoleWrite("Error in column name or index.")
        Case $ERROR_ERRUNIT
            ConsoleWrite("Error in unit name or index.")
        Case $ERROR_ERRZEROCOLUMNS
            ConsoleWrite("Error retrieving columns info, zero columns was returned.")
        Case $ERROR_ERRVALUE,
            ConsoleWrite("Plugin error when retrieving value.")
        Case $ERROR_ERRCALL
            ConsoleWrite("Error calling plugin function.")
        Case Else
            ConsoleWrite("Unknown error when retieving value.")
    EndSwitch
EndIf

_WriteOutput($g_hOutHandle, $sValue, $iCodepage)


Func ExitFunc()
    _WDXPluginClose($aPlugin)
    DllClose($g_hKernel32)
EndFunc

Func _GetTypeString($nType)
    If $nType > UBound($aTypes) Then $nType = 0
    Return $aTypes[$nType]
EndFunc

Func _TextAlign($sString, $nMaxCount)
    Local $sSpaces = "          "
    Local $nLen = StringLen($sString)
    If $nLen < $nMaxCount Then
        For $i = 1 To Ceiling(($nMaxCount - $nLen) / 10)
            $sString &= $sSpaces
        Next
    EndIf
    Return StringLeft($sString, $nMaxCount)
EndFunc

Func _WriteOutput($hHandle, $sText, $iCodepage = 65001)
    Local $aResult = DllCall($g_hKernel32, "int", "WideCharToMultiByte", _
            "uint", $iCodepage, _
            "dword", 0, _
            "wstr", $sText, _
            "int", -1, _
            "ptr", 0, _
            "int", 0, _
            "ptr", 0, _
            "ptr", 0)
    If @error Or Not $aResult[0] Then Return SetError(@error + 20, @extended, "")

    Local $nMultiByteSize = $aResult[0]
    Local $tMultiByte = DllStructCreate("byte[" & $nMultiByteSize & "]")

    $aResult = DllCall($g_hKernel32, "int", "WideCharToMultiByte", _
            "uint", $iCodepage, _
            "dword", 0, _
            "wstr", $sText, _
            "int", -1, _
            "struct*", $tMultiByte, _
            "int", $nMultiByteSize, _
            "ptr", 0, _
            "ptr", 0)
    If @error Or Not $aResult[0] Then Return SetError(@error + 10, @extended, "")

    Local $pNumberOfCharsWritten = 0
    ; $nMultiByteSize - 1, strip termination zero
    DllCall($g_hKernel32, "bool", "WriteFile", _
            "handle", $hHandle, _
            "ptr", DllStructGetPtr($tMultiByte), _
            "dword", $nMultiByteSize - 1, _
            "dword*", $pNumberOfCharsWritten, _
            "ptr", 0)
EndFunc

Func _GetStdHandle($nIndex)
    Local $aRes = DllCall($g_hKernel32, "handle", "GetStdHandle", "dword", $nIndex)
    Return $aRes[0]
EndFunc

Func _GetFullPath($sFileName)
    Local $tBuffer = DllStructCreate("wchar[4096]")
    ; Important, buffer size in TCHARs, or function fail 
    Local $aRet = DllCall("kernel32.dll", "dword", "GetFullPathNameW", _
                                          "wstr", $sFileName, _
                                          "dword", DllStructGetSize($tBuffer) / 2, _ 
                                          "ptr", DllStructGetPtr($tBuffer), _
                                          "ptr", 0)
    Return DllStructGetData($tBuffer, 1)
EndFunc